"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.hookArgs = void 0;
const config_js_1 = require("./config.js");
const interfaces_js_1 = require("./interfaces.js");
const utils_js_1 = require("./utils.js");
/**
 * Applies Graphile Config hooks to your GraphQL request, e.g. to
 * populate context or similar.
 *
 * @experimental
 */
function hookArgs(rawArgs, resolvedPreset, ctx) {
    const args = rawArgs;
    // Assert that args haven't already been hooked
    if (args[interfaces_js_1.$$hooked]) {
        throw new Error("Must not call hookArgs twice!");
    }
    args[interfaces_js_1.$$hooked] = true;
    // Make context mutable
    args.contextValue = Object.assign(Object.create(null), args.contextValue);
    // finalize(args): args is deliberately shadowed
    const finalize = (args) => {
        const userContext = resolvedPreset.grafast?.context;
        if (typeof userContext === "function") {
            const result = userContext(ctx, args);
            if ((0, utils_js_1.isPromiseLike)(result)) {
                // Deliberately shadowed 'result'
                return result.then((result) => {
                    Object.assign(args.contextValue, result);
                    return args;
                });
            }
            else {
                Object.assign(args.contextValue, result);
                return args;
            }
        }
        else if (typeof userContext === "object" && userContext !== null) {
            Object.assign(args.contextValue, userContext);
            return args;
        }
        else {
            return args;
        }
    };
    if (resolvedPreset !== config_js_1.NULL_PRESET &&
        resolvedPreset.plugins &&
        resolvedPreset.plugins.length > 0) {
        const event = { args, ctx, resolvedPreset };
        const result = (0, config_js_1.hook)(resolvedPreset, "args", event);
        if ((0, utils_js_1.isPromiseLike)(result)) {
            return result.then(() => finalize(event.args));
        }
        else {
            return finalize(event.args);
        }
    }
    return finalize(args);
}
exports.hookArgs = hookArgs;
//# sourceMappingURL=args.js.map