"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.hook = exports.withHooks = exports.withGraphileConfig = exports.NULL_PRESET = void 0;
exports.NULL_PRESET = Object.freeze(Object.create(null));
let graphileConfig = undefined;
let graphileConfigLoaded = false;
function withGraphileConfig(callback) {
    if (graphileConfig === undefined) {
        // ESM:
        // This should be
        //     graphileConfig = import("graphile-config").then(
        // but that causes a segfault in jest/node when testing third party
        // modules. So we had to convert everything to CommonJS.
        graphileConfig = new Promise((resolve, reject) => {
            try {
                resolve(require("graphile-config"));
            }
            catch (e) {
                if (e.code === "ERR_REQUIRE_ESM") {
                    return import("graphile-config").then(resolve, reject);
                }
                else {
                    reject(e);
                }
            }
        }).then((GC) => {
            graphileConfig = GC;
            graphileConfigLoaded = true;
            return GC;
        }, () => {
            graphileConfig = null;
            graphileConfigLoaded = true;
            return null;
        });
    }
    if (graphileConfigLoaded) {
        return callback(graphileConfig);
    }
    else {
        return graphileConfig.then(callback);
    }
}
exports.withGraphileConfig = withGraphileConfig;
const $$skipHooks = Symbol("skipHooks");
const $$hooksForPreset = Symbol("grafastHooks");
function withHooks(resolvedPreset, callback) {
    const existing = resolvedPreset[$$hooksForPreset];
    if (existing !== undefined) {
        return callback(existing);
    }
    if (!resolvedPreset.plugins || resolvedPreset.plugins.length === 0) {
        resolvedPreset[$$hooksForPreset] = null;
        return callback(null);
    }
    const plugins = resolvedPreset.plugins;
    return withGraphileConfig((gc) => {
        if (gc !== null) {
            const hooks = new gc.AsyncHooks();
            gc.applyHooks(plugins, (p) => p.grafast?.hooks, (name, fn, _plugin) => {
                hooks.hook(name, fn);
            });
            resolvedPreset[$$hooksForPreset] = hooks;
            return callback(hooks);
        }
        else {
            resolvedPreset[$$hooksForPreset] = null;
            return callback(null);
        }
    });
}
exports.withHooks = withHooks;
function hook(resolvedPreset, hookName, ...args) {
    if (resolvedPreset[$$skipHooks]?.[hookName]) {
        return;
    }
    return withHooks(resolvedPreset, (hooks) => {
        if (hooks !== null) {
            if (hooks.callbacks[hookName] !== undefined) {
                return hooks.process(hookName, ...args);
            }
            else {
                if (!resolvedPreset[$$skipHooks]) {
                    resolvedPreset[$$skipHooks] = Object.create(null);
                }
                resolvedPreset[$$skipHooks][hookName] = true;
                return;
            }
        }
        else {
            if (!resolvedPreset[$$skipHooks]) {
                resolvedPreset[$$skipHooks] = Object.create(null);
            }
            resolvedPreset[$$skipHooks][hookName] = true;
            return;
        }
    });
}
exports.hook = hook;
//# sourceMappingURL=config.js.map