"use strict";
// PERF: optimise constraint matching, e.g. by collapsing redundant constraints
// and then compiling (JIT-style) remaining constraints into a function using
// tamedevil
Object.defineProperty(exports, "__esModule", { value: true });
exports.matchesConstraints = void 0;
function valueAtPath(object, path) {
    let value = object;
    for (let i = 0, l = path.length; i < l; i++) {
        const key = path[i];
        const expectArray = typeof key === "number";
        const isArray = Array.isArray(value);
        if (expectArray !== isArray) {
            return undefined;
        }
        else {
            if (value == null) {
                return undefined;
            }
            value = value[key];
            if (typeof value === "undefined") {
                return undefined;
            }
        }
    }
    return value;
}
/**
 * Implements the `MatchesConstraint` algorithm.
 */
function matchesConstraint(constraint, object) {
    const value = valueAtPath(object, constraint.path);
    switch (constraint.type) {
        case "length": {
            return Array.isArray(value) && value.length === constraint.expectedLength;
        }
        case "exists": {
            return (value !== undefined) === constraint.exists;
        }
        case "equal": {
            return (value === constraint.expectedValue) === constraint.pass;
        }
        case "value": {
            return value === constraint.value;
        }
        case "isEmpty": {
            const isEmpty = typeof value === "object" &&
                value !== null &&
                Object.keys(value).length === 0;
            return isEmpty === constraint.isEmpty;
        }
        default: {
            const never = constraint;
            throw new Error(`Unsupported constraint type '${never.type}'`);
        }
    }
}
/**
 * Implements the `MatchesConstraints` algorithm.
 */
function matchesConstraints(constraints, object) {
    // In my testing, verbose loops are still about 20% faster than for...of
    for (let i = 0, l = constraints.length; i < l; i++) {
        const constraint = constraints[i];
        if (!matchesConstraint(constraint, object)) {
            return false;
        }
    }
    return true;
}
exports.matchesConstraints = matchesConstraints;
//# sourceMappingURL=constraints.js.map