"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.execute = exports.withGrafastArgs = void 0;
const tslib_1 = require("tslib");
const eventemitter3_1 = tslib_1.__importDefault(require("eventemitter3"));
const config_js_1 = require("./config.js");
const dev_js_1 = require("./dev.js");
const inspect_js_1 = require("./inspect.js");
const interfaces_js_1 = require("./interfaces.js");
const prepare_js_1 = require("./prepare.js");
const utils_js_1 = require("./utils.js");
/**
 * Used by `execute` and `subscribe`.
 * @internal
 */
function withGrafastArgs(args, resolvedPreset, outputDataAsString) {
    const options = resolvedPreset?.grafast;
    if (dev_js_1.isDev) {
        if (args.rootValue != null &&
            (typeof args.rootValue !== "object" ||
                Object.keys(args.rootValue).length > 0)) {
            throw new Error(`Grafast executor doesn't support there being a rootValue (found ${(0, inspect_js_1.inspect)(args.rootValue)})`);
        }
    }
    if (args.rootValue == null) {
        args.rootValue = Object.create(null);
    }
    if (typeof args.rootValue !== "object" || args.rootValue == null) {
        throw new Error("Grafast requires that the 'rootValue' be an object");
    }
    const explain = options?.explain;
    const shouldExplain = !!explain;
    let unlisten = null;
    if (shouldExplain) {
        const eventEmitter = new eventemitter3_1.default();
        const explainOperations = [];
        args.rootValue = Object.assign(Object.create(null), args.rootValue, {
            [interfaces_js_1.$$eventEmitter]: eventEmitter,
            [interfaces_js_1.$$extensions]: {
                explain: {
                    operations: explainOperations,
                },
            },
        });
        const handleExplainOperation = ({ operation, }) => {
            if (explain === true || (explain && explain.includes(operation.type))) {
                explainOperations.push(operation);
            }
        };
        eventEmitter.on("explainOperation", handleExplainOperation);
        unlisten = () => {
            eventEmitter.removeListener("explainOperation", handleExplainOperation);
        };
    }
    const rootValue = (0, prepare_js_1.grafastPrepare)(args, {
        explain: options?.explain,
        outputDataAsString,
        timeouts: options?.timeouts,
    });
    if (unlisten !== null) {
        Promise.resolve(rootValue).then(unlisten, unlisten);
    }
    // Convert from PromiseOrDirect to PromiseOrValue
    if ((0, utils_js_1.isPromiseLike)(rootValue)) {
        return Promise.resolve(rootValue);
    }
    else {
        return rootValue;
    }
}
exports.withGrafastArgs = withGrafastArgs;
function execute(args, resolvedPreset, outputDataAsString) {
    return withGrafastArgs(args, args.resolvedPreset ?? resolvedPreset ?? config_js_1.NULL_PRESET, args.outputDataAsString ?? outputDataAsString ?? false);
}
exports.execute = execute;
//# sourceMappingURL=execute.js.map