"use strict";
var _a, _b;
Object.defineProperty(exports, "__esModule", { value: true });
exports.assertListCapableStep = exports.isListCapableStep = exports.assertModifierStep = exports.isModifierStep = exports.ModifierStep = exports.isPolymorphicStep = exports.isStreamableStep = exports.isListLikeStep = exports.isObjectLikeStep = exports.isUnbatchedExecutableStep = exports.assertExecutableStep = exports.isExecutableStep = exports.UnbatchedExecutableStep = exports.ExecutableStep = exports.BaseStep = exports.assertFinalized = exports.$$noExec = void 0;
const tslib_1 = require("tslib");
const chalk_1 = tslib_1.__importDefault(require("chalk"));
const tamedevil_1 = tslib_1.__importDefault(require("tamedevil"));
const dev_js_1 = require("./dev.js");
const withGlobalLayerPlan_js_1 = require("./engine/lib/withGlobalLayerPlan.js");
const lock_js_1 = require("./engine/lock.js");
const error_js_1 = require("./error.js");
const global_js_1 = require("./global.js");
const inspect_js_1 = require("./inspect.js");
const interfaces_js_1 = require("./interfaces.js");
const utils_js_1 = require("./utils.js");
/**
 * This indicates that a step never executes (e.g. __ItemStep and __ValueStep)
 * and thus when executed skips direct to reallyCompletedStep.
 *
 * @internal
 */
exports.$$noExec = Symbol("noExec");
const ref_flagError = tamedevil_1.default.ref(error_js_1.flagError, "flagError");
function throwDestroyed() {
    let message;
    try {
        message = `${this} has been destroyed; calling methods on it is no longer possible`;
    }
    catch (e) {
        message = `Step ${this?.id} has been destroyed; calling methods on it is no longer possible`;
    }
    throw new Error(message);
}
function isDeepDepSkippable($dep) {
    return interfaces_js_1.$$deepDepSkip in $dep && typeof $dep[interfaces_js_1.$$deepDepSkip] === "function";
}
function reallyAssertFinalized(plan) {
    if (!plan.isFinalized) {
        throw new Error(`Step ${plan} is not finalized; did you forget to call \`super.finalize()\` from its \`finalize()\` method?`);
    }
}
// Optimise this away in production.
exports.assertFinalized = !dev_js_1.isDev ? dev_js_1.noop : reallyAssertFinalized;
/**
 * The base abstract plan type; you should not extend this directly - instead
 * use an ExecutableStep (for use when planning output fields) or a
 * ModifierStep (for use when planning arguments/input fields).
 *
 * @remarks
 *
 * Though it might seem that ModifierStep should be used for all inputs
 * (arguments, input objects), this is not the case. When planning an output
 * field, all inputs to it (including arguments, parent plan, context, etc)
 * should be other executable plans. The only time that ModifierStep is used is
 * when writing specific plans as part of the argument or input field
 * definitions themself; even in these cases the inputs to these plan resolvers
 * will be ExecutablePlans.
 */
class BaseStep {
    static { _a = interfaces_js_1.$$subroutine; }
    constructor() {
        this[_a] = null;
        this.isArgumentsFinalized = false;
        this.isFinalized = false;
        this.debug = (0, global_js_1.getDebug)();
        this.hasSideEffects = false;
        const layerPlan = (0, withGlobalLayerPlan_js_1.currentLayerPlan)();
        this.layerPlan = layerPlan;
        this.operationPlan = layerPlan.operationPlan;
        this._isUnary = true;
        this._isUnaryLocked = false;
    }
    toString() {
        const meta = this.toStringMeta();
        return chalk_1.default.bold.blue(`${this.constructor.name.replace(/Step$/, "")}${meta != null && meta.length ? chalk_1.default.grey(`<${meta}>`) : ""}`);
    }
    /**
     * This metadata will be merged into toString when referencing this plan.
     */
    toStringMeta() {
        return null;
    }
    planJSONExtra() {
        return undefined;
    }
    finalize() {
        if (!this.isFinalized) {
            this.isFinalized = true;
        }
        else {
            throw new Error(`Step ${this} has already been finalized - do not call \`finalize()\` from user code!`);
        }
    }
    destroy() { }
}
exports.BaseStep = BaseStep;
/**
 * Executable plans are the plans associated with leaves on the GraphQL tree,
 * they must be able to execute to return values.
 */
class ExecutableStep extends BaseStep {
    static { _b = lock_js_1.$$unlock; }
    constructor() {
        super();
        /** @internal */
        this[_b] = undefined;
        /**
         * (default = ALL_FLAGS & ~FLAG_NULL)
         */
        this.defaultForbiddenFlags = interfaces_js_1.DEFAULT_FORBIDDEN_FLAGS;
        this.dependencies = [];
        this.dependencyForbiddenFlags = [];
        this.dependencyOnReject = [];
        this.dependents = [];
        this.isOptimized = false;
        this.allowMultipleOptimizations = false;
        this._stepOptions = { stream: null };
        this.store = true;
        this.polymorphicPaths = (0, withGlobalLayerPlan_js_1.currentPolymorphicPaths)();
        this.id = this.layerPlan._addStep(this);
    }
    withMyLayerPlan(callback) {
        return (0, withGlobalLayerPlan_js_1.withGlobalLayerPlan)(this.layerPlan, this.polymorphicPaths, callback);
    }
    getStep(id) {
        return this.layerPlan.getStep(id, this);
    }
    getDepOptions(depId) {
        const step = this.dependencies[depId];
        const forbiddenFlags = this.dependencyForbiddenFlags[depId];
        const onReject = this.dependencyOnReject[depId];
        const acceptFlags = interfaces_js_1.ALL_FLAGS & ~forbiddenFlags;
        return { step, acceptFlags, onReject };
    }
    getDep(_depId) {
        // This gets replaced when `__FlagStep` is loaded. Were we on ESM we could
        // just put the code here, but since we're not we have to avoid the
        // circular dependency.
        throw new Error(`Grafast failed to load correctly`);
    }
    /**
     * Like getDep, except it skips over __ItemStep and similar steps to get to
     * where the parent really is.
     *
     * @experimental
     */
    getDepDeep(depId) {
        let $dep = this.getDep(depId);
        // Walk up the tree, looking for the source of this record. We know that
        // __ItemStep and __ListTransformStep are safe to walk through, but other
        // classes may not be.
        while (isDeepDepSkippable($dep)) {
            $dep = $dep[interfaces_js_1.$$deepDepSkip]();
        }
        return $dep;
    }
    toString() {
        let meta;
        try {
            // If we log out too early, the meta function might fail.
            meta = this.toStringMeta();
        }
        catch (e) {
            // Ignore
        }
        return chalk_1.default.bold.blue(`${this.constructor.name.replace(/Step$/, "")}${this.layerPlan.id === 0 ? "" : chalk_1.default.grey(`{${this.layerPlan.id}}`)}${meta != null && meta.length ? chalk_1.default.grey(`<${meta}>`) : ""}[${(0, inspect_js_1.inspect)(this.id, {
            colors: true,
        })}]`);
    }
    canAddDependency(step) {
        return (0, utils_js_1.stepAMayDependOnStepB)(this, step);
    }
    addDependency(stepOrOptions) {
        const options = stepOrOptions instanceof ExecutableStep
            ? { step: stepOrOptions, skipDeduplication: false }
            : stepOrOptions;
        return this.operationPlan.stepTracker.addStepDependency(this, options);
    }
    /**
     * Adds "unary" dependencies; in `execute({count, values})` you'll receive a
     * `values[index]` (where `index` is the return value of this function) with
     * `isBatch = false` so you can use the `values[index].value` property
     * directly.
     */
    addUnaryDependency(stepOrOptions) {
        const options = stepOrOptions instanceof ExecutableStep
            ? { step: stepOrOptions }
            : stepOrOptions;
        return this.operationPlan.stepTracker.addStepUnaryDependency(this, options);
    }
    finalize() {
        if (typeof this.isSyncAndSafe !== "boolean") {
            // Take a guess
            if (!dev_js_1.isDev) {
                this.isSyncAndSafe = false;
            }
            else if (this.execute.constructor.name === "AsyncFunction") {
                this.isSyncAndSafe = false;
            }
            else {
                console.warn(`${this} uses a regular (non-async) function for 'execute'; if it never returns a promise and the list it returns never includes a promise then setting \`${this}.isSyncAndSafe = true\` will improve performance. If this is not true, set \`${this}.isSyncAndSafe = false\` to dismiss this message.`);
                this.isSyncAndSafe = false;
            }
        }
        if (this.isSyncAndSafe === true &&
            this.execute.constructor.name === "AsyncFunction") {
            throw new Error(`${this} claims to be synchronous, however the execute method is asynchronous`);
        }
        super.finalize();
    }
    /**
     * This function will be called with 'execution details', an object containing:
     *
     * - `count`: the number of entries in the batch that's being executed
     * - `values`: a tuple representing the runtime values of the steps
     *   dependencies; each value in the tuple is an object, either a batch object
     *   containing a list of size `count` containing the values, or a unary
     *   object containing the single value common to all entries.
     * - `indexMap`: helper function to map over each index from `0` to `count-1`,
     *   returning the resulting array.
     * - `indexForEach`: as `indexMap`, but without the array result.
     * - `meta`: [experimental]
     *
     * `execute` must return a list with `count` entries, where each value in the
     * list relates to the result of executing this plan for the corresponding
     * entry in each of the entries in the `values` tuple.
     *
     * IMPORTANT: it is up to the execute function to cache/memoize results as
     * appropriate for performance, this can be done via the `meta` object.
     *
     * The `meta` object is an empty object stored to `grafastContext.metaByPlan`
     * that can be used to store anything this plan needs. We recommend that you
     * add attributes to meta for each purpose (e.g. use `meta.cache` for
     * memoizing results) so that you can expand your usage of meta in future.
     */
    /* abstract */
    execute(details) {
        // ESLint/TS: ignore not used.
        details;
        throw new Error(`${this} has not implemented an 'execute' method`);
    }
    destroy() {
        // Break ourself enough that if lifecycle methods are attempted an error
        // will be thrown. This should help weed out bugs where steps are processed
        // even after they have been removed/deduped.
        this.addDependency = throwDestroyed;
        this.deduplicate = throwDestroyed;
        this.deduplicatedWith = throwDestroyed;
        this.optimize = throwDestroyed;
        this.finalize = throwDestroyed;
        this.execute = throwDestroyed;
        super.destroy();
    }
}
exports.ExecutableStep = ExecutableStep;
function _buildOptimizedExecuteV2Expression(depCount, isSyncAndSafe) {
    const identifiers = [];
    for (let i = 0; i < depCount; i++) {
        identifiers.push(tamedevil_1.default.identifier(`value${i}`));
    }
    const tryOrNot = (inFrag) => {
        if (isSyncAndSafe) {
            return inFrag;
        }
        else {
            return (0, tamedevil_1.default) `\
    try {
  ${tamedevil_1.default.indent(inFrag)}
    } catch (e) {
      results[i] = ${ref_flagError}(e);
    }\
`;
        }
    };
    return (0, tamedevil_1.default) `\
(function execute({
  count,
  values: [${tamedevil_1.default.join(identifiers, ", ")}],
  extra,
}) {
  const results = [];
  for (let i = 0; i < count; i++) {
${tryOrNot((0, tamedevil_1.default) `\
    results[i] = this.unbatchedExecute(extra, ${tamedevil_1.default.join(identifiers.map((identifier) => (0, tamedevil_1.default) `${identifier}.at(i)`), ", ")});\
`)}
  }
  return results;
})`;
}
const MAX_DEPENDENCIES_TO_CACHE = 10;
const unsafeCache = [];
const safeCache = [];
tamedevil_1.default.batch(() => {
    for (let i = 0; i <= MAX_DEPENDENCIES_TO_CACHE; i++) {
        const depCount = i;
        const unsafeExpression = _buildOptimizedExecuteV2Expression(depCount, false);
        tamedevil_1.default.runInBatch(unsafeExpression, (fn) => {
            unsafeCache[depCount] = fn;
        });
        const safeExpression = _buildOptimizedExecuteV2Expression(depCount, true);
        tamedevil_1.default.runInBatch(safeExpression, (fn) => {
            safeCache[depCount] = fn;
        });
    }
});
function buildOptimizedExecute(depCount, isSyncAndSafe, callback) {
    // Try and satisfy from cache
    const cache = isSyncAndSafe ? safeCache : unsafeCache;
    if (depCount <= MAX_DEPENDENCIES_TO_CACHE) {
        callback(cache[depCount]);
        return;
    }
    // Build it
    const expression = _buildOptimizedExecuteV2Expression(depCount, isSyncAndSafe);
    tamedevil_1.default.runInBatch(expression, (fn) => {
        callback(fn);
    });
}
class UnbatchedExecutableStep extends ExecutableStep {
    static { this.$$export = {
        moduleName: "grafast",
        exportName: "UnbatchedExecutableStep",
    }; }
    finalize() {
        if (this.execute === UnbatchedExecutableStep.prototype.execute) {
            // If they've not replaced 'execute', use our optimized form
            buildOptimizedExecute(this.dependencies.length, this.isSyncAndSafe, (fn) => {
                this.execute = fn;
            });
        }
        super.finalize();
    }
    execute({ indexMap, values, extra, }) {
        console.warn(`${this} didn't call 'super.finalize()' in the finalize method.`);
        return indexMap((i) => {
            try {
                const tuple = values.map((list) => list.at(i));
                return this.unbatchedExecute(extra, ...tuple);
            }
            catch (e) {
                return (0, error_js_1.flagError)(e);
            }
        });
    }
}
exports.UnbatchedExecutableStep = UnbatchedExecutableStep;
function isExecutableStep(step) {
    return (step instanceof BaseStep &&
        "execute" in step &&
        typeof step.execute === "function");
}
exports.isExecutableStep = isExecutableStep;
function assertExecutableStep(step) {
    if (!isExecutableStep(step)) {
        throw new Error(`Expected an executable step, but received something else: ${(0, inspect_js_1.inspect)(step)}`);
    }
}
exports.assertExecutableStep = assertExecutableStep;
function isUnbatchedExecutableStep(step) {
    return (isExecutableStep(step) &&
        typeof step.unbatchedExecute === "function");
}
exports.isUnbatchedExecutableStep = isUnbatchedExecutableStep;
function isObjectLikeStep(plan) {
    return "get" in plan && typeof plan.get === "function";
}
exports.isObjectLikeStep = isObjectLikeStep;
function isListLikeStep(plan) {
    return "at" in plan && typeof plan.at === "function";
}
exports.isListLikeStep = isListLikeStep;
function isStreamableStep(plan) {
    return typeof plan.stream === "function";
}
exports.isStreamableStep = isStreamableStep;
function isPolymorphicStep(s) {
    return ("planForType" in s &&
        typeof s.planForType === "function");
}
exports.isPolymorphicStep = isPolymorphicStep;
/**
 * Modifier plans modify their parent plan (which may be another ModifierStep
 * or an ExecutableStep). First they gather all the requirements from their
 * children (if any) being applied to them, then they apply themselves to their
 * parent plan. This application is done through the `apply()` method.
 *
 * Modifier plans do not use dependencies.
 */
class ModifierStep extends BaseStep {
    constructor($parent) {
        super();
        this.$parent = $parent;
        this.id = this.layerPlan._addModifierStep(this);
    }
    toString() {
        const meta = this.toStringMeta();
        return chalk_1.default.bold.blue(`${this.constructor.name.replace(/Step$/, "")}${meta != null && meta.length ? chalk_1.default.grey(`<${meta}>`) : ""}[${(0, inspect_js_1.inspect)(this.id, {
            colors: true,
        })}]`);
    }
}
exports.ModifierStep = ModifierStep;
function isModifierStep(plan) {
    return "apply" in plan && typeof plan.apply === "function";
}
exports.isModifierStep = isModifierStep;
function assertModifierStep(plan, pathDescription) {
    if (!isModifierStep(plan)) {
        throw new Error(`The plan returned from '${pathDescription}' should be a modifier plan, but it does not implement the 'apply' method.`);
    }
}
exports.assertModifierStep = assertModifierStep;
function isListCapableStep(plan) {
    return "listItem" in plan && typeof plan.listItem === "function";
}
exports.isListCapableStep = isListCapableStep;
function assertListCapableStep(plan, pathDescription) {
    if (!isListCapableStep(plan)) {
        throw new Error(`The plan returned from '${pathDescription}' should be a list capable plan, but ${plan} does not implement the 'listItem' method.`);
    }
}
exports.assertListCapableStep = assertListCapableStep;
//# sourceMappingURL=step.js.map