import type { ExecutionDetails, ExecutionEntryFlags, GrafastResultsList } from "../interfaces.js";
import { $$deepDepSkip } from "../interfaces.js";
import { ExecutableStep } from "../step.js";
import type { __ItemStep } from "./__item.js";
export declare const TRAP_ERROR: ExecutionEntryFlags;
export declare const TRAP_INHIBITED: ExecutionEntryFlags;
export declare const TRAP_ERROR_OR_INHIBITED: ExecutionEntryFlags;
declare const TRAP_VALUES: readonly ["NULL", "EMPTY_LIST", "PASS_THROUGH"];
/** @defaultValue `'PASS_THROUGH'` */
export type TrapValue = (typeof TRAP_VALUES)[number];
/** `false` means pass-through; all others are literal */
export type ResolvedTrapValue = false | null | undefined | readonly never[];
export interface FlagStepOptions {
    acceptFlags?: ExecutionEntryFlags;
    onReject?: Error | null;
    if?: ExecutableStep<boolean>;
    valueForInhibited?: TrapValue;
    valueForError?: TrapValue;
}
export declare class __FlagStep<TData> extends ExecutableStep<TData> {
    static $$export: {
        moduleName: string;
        exportName: string;
    };
    isSyncAndSafe: boolean;
    private ifDep;
    private forbiddenFlags;
    private onRejectReturnValue;
    private valueForInhibited;
    private valueForError;
    private canBeInlined;
    constructor(step: ExecutableStep, options: FlagStepOptions);
    toStringMeta(): string | null;
    [$$deepDepSkip](): ExecutableStep;
    listItem?: ($item: __ItemStep<this>) => ExecutableStep;
    _listItem($item: __ItemStep<this>): ExecutableStep<unknown>;
    deduplicate(_peers: readonly ExecutableStep<any>[]): readonly ExecutableStep<any>[];
    execute(_details: ExecutionDetails<[data: TData, cond?: boolean]>): GrafastResultsList<TData>;
    finalize(): void;
    private fancyExecute;
    private passThroughExecute;
}
/**
 * Example use case: get user by id, but id is null: no need to fetch the user
 * since we know they won't exist.
 */
export declare function inhibitOnNull<T>($step: ExecutableStep<T>): __FlagStep<T>;
/**
 * Example use case: expecting a node ID that represents a User, but get one
 * that represents a Post instead: throw error to tell user they've sent invalid
 * data.
 */
export declare function assertNotNull<T>($step: ExecutableStep<T>, message: string, options?: {
    if?: FlagStepOptions["if"];
}): __FlagStep<T>;
export declare function trap<T>($step: ExecutableStep<T>, acceptFlags: ExecutionEntryFlags, options?: {
    valueForInhibited?: FlagStepOptions["valueForInhibited"];
    valueForError?: FlagStepOptions["valueForError"];
}): __FlagStep<T>;
export {};
//# sourceMappingURL=__flag.d.ts.map