"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.constant = exports.ConstantStep = void 0;
const inspect_js_1 = require("../inspect.js");
const step_js_1 = require("../step.js");
const utils_js_1 = require("../utils.js");
/**
 * Converts a constant value (e.g. a string/number/etc) into a plan
 */
class ConstantStep extends step_js_1.UnbatchedExecutableStep {
    static { this.$$export = {
        moduleName: "grafast",
        exportName: "ConstantStep",
    }; }
    constructor(data, isSensitive = typeof data !== "boolean" && data != null) {
        super();
        this.data = data;
        this.isSensitive = isSensitive;
        this.isSyncAndSafe = true;
    }
    toStringMeta() {
        // ENHANCE: use nicer simplification
        return this.isSensitive
            ? `[HIDDEN]`
            : (0, inspect_js_1.inspect)(this.data)
                .replace(/[\r\n]/g, " ")
                .slice(0, 60);
    }
    deduplicate(peers) {
        return peers.filter((p) => p.data === this.data);
    }
    execute({ count }) {
        return (0, utils_js_1.arrayOfLength)(count, this.data);
    }
    eval() {
        return this.data;
    }
    evalIs(value) {
        return this.data === value;
    }
    evalIsEmpty() {
        return (typeof this.data === "object" &&
            this.data !== null &&
            Object.keys(this.data).length === 0);
    }
    unbatchedExecute() {
        return this.data;
    }
    isNull() {
        return this.data === null;
    }
    isUndefined() {
        return this.data === undefined;
    }
    get(key) {
        const value = typeof this.data === "object" &&
            this.data !== null &&
            Object.hasOwn(this.data, key)
            ? this.data[key]
            : undefined;
        return constant(value);
    }
    at(index) {
        const value = Array.isArray(this.data) ? this.data[index] : undefined;
        return constant(value);
    }
}
exports.ConstantStep = ConstantStep;
function isTemplateStringsArray(data) {
    return (Array.isArray(data) &&
        "raw" in data &&
        Array.isArray(data.raw));
}
function constant(data, isSecret) {
    if (isTemplateStringsArray(data)) {
        if (data.length !== 1) {
            throw new Error("constant`...` doesn't currently support placeholders; please use 'constant(`...`)' instead");
        }
        return new ConstantStep(data[0], false);
    }
    return new ConstantStep(data, isSecret);
}
exports.constant = constant;
//# sourceMappingURL=constant.js.map