"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.list = exports.ListStep = void 0;
const step_js_1 = require("../step.js");
const constant_js_1 = require("./constant.js");
class ListStep extends step_js_1.UnbatchedExecutableStep {
    static { this.$$export = {
        moduleName: "grafast",
        exportName: "ListStep",
    }; }
    constructor(list) {
        super();
        this.isSyncAndSafe = true;
        this.allowMultipleOptimizations = true;
        this.optimizeMetaKey = "ListStep";
        for (let i = 0, l = list.length; i < l; i++) {
            this.addDependency({ step: list[i], skipDeduplication: true });
        }
    }
    toStringMeta() {
        return this.dependencies.map(($dep) => $dep.id).join(",");
    }
    execute({ indexMap, values }) {
        return indexMap((i) => values.map((value) => value.at(i)));
    }
    unbatchedExecute(_extra, ...values //UnwrapPlanTuple<TPlanTuple>,
    ) {
        return values;
    }
    deduplicate(peers) {
        return peers;
    }
    optimize(opts) {
        if (this.dependencies.every((dep) => dep instanceof constant_js_1.ConstantStep)) {
            const meta = opts.meta;
            // Used to validate the lists have the same length
            const cardinality = this.dependencies.length;
            if (!meta[cardinality]) {
                meta[cardinality] = [];
            }
            const existing = meta[cardinality].find((l) => l.every((v, i) => v === this.dependencies[i].data));
            const isSensitive = this.dependencies.some((d) => d.isSensitive);
            if (existing !== undefined) {
                return (0, constant_js_1.constant)(existing, isSensitive);
            }
            else {
                // Replace self with constant
                const arr = this.dependencies.map((dep) => dep.data);
                meta[cardinality].push(arr);
                return (0, constant_js_1.constant)(arr, isSensitive);
            }
        }
        return this;
    }
    /**
     * Get the original plan at the given index back again.
     */
    at(index) {
        return this.getDep(index);
    }
    first() {
        return this.getDep(0);
    }
    last() {
        return this.getDep(this.dependencies.length - 1);
    }
}
exports.ListStep = ListStep;
/**
 * Takes a list of plans and turns it into a single plan that represents the
 * list of their values.
 */
function list(list) {
    return new ListStep(list);
}
exports.list = list;
//# sourceMappingURL=list.js.map