import type { __ItemStep, ExecutionDetails } from "../index.js";
import type { GrafastResultsList, Maybe, PromiseOrDirect } from "../interfaces.js";
import { ExecutableStep } from "../step.js";
export interface LoadOptions<TItem, TParams extends Record<string, any>, TUnarySpec = never> {
    unary: TUnarySpec;
    attributes: ReadonlyArray<keyof TItem>;
    params: Partial<TParams>;
}
type LoadCallback<TSpec, TItem, TData extends TItem | ReadonlyArray<TItem>, TParams extends Record<string, any>, TUnarySpec = never> = {
    (specs: ReadonlyArray<TSpec>, options: LoadOptions<TItem, TParams, TUnarySpec>): PromiseOrDirect<ReadonlyArray<Maybe<TData>>>;
    displayName?: string;
};
export type LoadOneCallback<TSpec, TItem, TParams extends Record<string, any>, TUnarySpec = never> = LoadCallback<TSpec, TItem, TItem, TParams, TUnarySpec>;
export type LoadManyCallback<TSpec, TItem, TParams extends Record<string, any>, TUnarySpec = never> = LoadCallback<TSpec, TItem, ReadonlyArray<TItem>, TParams, TUnarySpec>;
/**
 * A TypeScript Identity Function to help you strongly type your
 * LoadOneCallback.
 */
export declare function loadOneCallback<TSpec, TItem, TParams extends Record<string, any>, TUnarySpec = never>(callback: LoadOneCallback<TSpec, TItem, TParams, TUnarySpec>): LoadOneCallback<TSpec, TItem, TParams, TUnarySpec>;
/**
 * A TypeScript Identity Function to help you strongly type your
 * LoadManyCallback.
 */
export declare function loadManyCallback<TSpec, TItem, TParams extends Record<string, any>, TUnarySpec = never>(callback: LoadManyCallback<TSpec, TItem, TParams, TUnarySpec>): LoadManyCallback<TSpec, TItem, TParams, TUnarySpec>;
/**
 * You shouldn't create instances of this yourself - use `loadOne` or `loadMany` instead.
 */
export declare class LoadedRecordStep<TItem, TParams extends Record<string, any>> extends ExecutableStep<TItem> {
    private isSingle;
    private sourceDescription;
    private ioEquivalence;
    static $$export: {
        moduleName: string;
        exportName: string;
    };
    isSyncAndSafe: boolean;
    attributes: Set<keyof TItem>;
    params: Partial<TParams>;
    constructor($data: ExecutableStep<TItem>, isSingle: boolean, sourceDescription: string, ioEquivalence: Record<string, ExecutableStep>);
    toStringMeta(): string;
    get(attr: keyof TItem & (string | number)): ExecutableStep<any>;
    setParam<TParamKey extends keyof TParams>(paramKey: TParamKey, value: TParams[TParamKey]): void;
    optimize(): ExecutableStep<any>;
    execute({ count, values: [values0], }: ExecutionDetails<[TItem]>): GrafastResultsList<TItem>;
}
export declare class LoadStep<TSpec, TItem, TData extends TItem | ReadonlyArray<TItem>, TParams extends Record<string, any>, TUnarySpec = never> extends ExecutableStep {
    private ioEquivalence;
    private load;
    static $$export: {
        moduleName: string;
        exportName: string;
    };
    isSyncAndSafe: boolean;
    loadOptions: Omit<LoadOptions<TItem, TParams, TUnarySpec>, "unary"> | null;
    loadOptionsKey: string;
    attributes: Set<keyof TItem>;
    params: Partial<TParams>;
    unaryDepId: number | null;
    constructor($spec: ExecutableStep<TSpec>, $unarySpec: ExecutableStep<TUnarySpec> | null, ioEquivalence: null | string | {
        [key in keyof TSpec]?: string | null;
    }, load: LoadCallback<TSpec, TItem, TData, TParams, TUnarySpec>);
    toStringMeta(): string;
    private makeAccessMap;
    listItem($item: __ItemStep<TItem>): LoadedRecordStep<TItem, TParams>;
    single(): TData extends ReadonlyArray<any> ? never : LoadedRecordStep<TItem, TParams>;
    setParam<TParamKey extends keyof TParams>(paramKey: TParamKey, value: TParams[TParamKey]): void;
    addAttributes(attributes: Set<keyof TItem>): void;
    finalize(): void;
    execute({ count, values: [values0, values1], extra, }: ExecutionDetails<[TSpec, TUnarySpec]>): PromiseOrDirect<GrafastResultsList<Maybe<TData>>>;
}
export declare function loadMany<TSpec, TItem, TParams extends Record<string, any> = Record<string, any>, TUnarySpec = never>($spec: ExecutableStep<TSpec>, loadCallback: LoadManyCallback<TSpec, TItem, TParams, TUnarySpec>): LoadStep<TSpec, TItem, ReadonlyArray<TItem>, TParams, TUnarySpec>;
export declare function loadMany<TSpec, TItem, TParams extends Record<string, any> = Record<string, any>, TUnarySpec = never>($spec: ExecutableStep<TSpec>, ioEquivalence: null | string | (TSpec extends [...any[]] ? {
    [key in keyof TSpec]: string | null;
} : TSpec extends Record<string, any> ? {
    [key in keyof TSpec]?: string | null;
} : never), loadCallback: LoadManyCallback<TSpec, TItem, TParams, TUnarySpec>): LoadStep<TSpec, TItem, ReadonlyArray<TItem>, TParams, TUnarySpec>;
export declare function loadMany<TSpec, TItem, TParams extends Record<string, any> = Record<string, any>, TUnarySpec = never>($spec: ExecutableStep<TSpec>, $unarySpec: ExecutableStep<TUnarySpec> | null, loadCallback: LoadManyCallback<TSpec, TItem, TParams, TUnarySpec>): LoadStep<TSpec, TItem, ReadonlyArray<TItem>, TParams, TUnarySpec>;
export declare function loadMany<TSpec, TItem, TParams extends Record<string, any> = Record<string, any>, TUnarySpec = never>($spec: ExecutableStep<TSpec>, $unarySpec: ExecutableStep<TUnarySpec> | null, ioEquivalence: null | string | (TSpec extends [...any[]] ? {
    [key in keyof TSpec]: string | null;
} : TSpec extends Record<string, any> ? {
    [key in keyof TSpec]?: string | null;
} : never), loadCallback: LoadManyCallback<TSpec, TItem, TParams, TUnarySpec>): LoadStep<TSpec, TItem, ReadonlyArray<TItem>, TParams, TUnarySpec>;
export declare function loadOne<const TSpec, TItem, TParams extends Record<string, any> = Record<string, any>, TUnarySpec = never>($spec: ExecutableStep<TSpec>, loadCallback: LoadOneCallback<TSpec, TItem, TParams, TUnarySpec>): LoadedRecordStep<TItem, TParams>;
export declare function loadOne<const TSpec, TItem, TParams extends Record<string, any> = Record<string, any>, TUnarySpec = never>($spec: ExecutableStep<TSpec>, ioEquivalence: null | string | {
    [key in keyof TSpec]?: string | null;
}, loadCallback: LoadOneCallback<TSpec, TItem, TParams, TUnarySpec>): LoadedRecordStep<TItem, TParams>;
export declare function loadOne<const TSpec, TItem, TParams extends Record<string, any> = Record<string, any>, const TUnarySpec = never>($spec: ExecutableStep<TSpec>, $unarySpec: ExecutableStep<TUnarySpec> | null, loadCallback: LoadOneCallback<TSpec, TItem, TParams, TUnarySpec>): LoadedRecordStep<TItem, TParams>;
export declare function loadOne<const TSpec, TItem, TParams extends Record<string, any> = Record<string, any>, const TUnarySpec = never>($spec: ExecutableStep<TSpec>, $unarySpec: ExecutableStep<TUnarySpec> | null, ioEquivalence: null | string | {
    [key in keyof TSpec]?: string | null;
}, loadCallback: LoadOneCallback<TSpec, TItem, TParams, TUnarySpec>): LoadedRecordStep<TItem, TParams>;
export {};
//# sourceMappingURL=load.d.ts.map