import type { DataFromStep, ExecutionDetails, ExecutionExtra, StepOptimizeOptions, UnbatchedExecutionExtra } from "../interfaces.js";
import type { ExecutableStep } from "../step.js";
import { UnbatchedExecutableStep } from "../step.js";
import { ConstantStep } from "./constant.js";
import type { SetterCapableStep } from "./setter.js";
export type DataFromObjectSteps<TSteps extends {
    [key: string]: ExecutableStep;
}> = {
    [key in keyof TSteps]: DataFromStep<TSteps[key]>;
};
type Results<TSteps extends {
    [key: string]: ExecutableStep;
}> = Array<[
    Array<DataFromObjectSteps<TSteps>[keyof TSteps]>,
    DataFromObjectSteps<TSteps>
]>;
export interface ObjectPlanMeta<TSteps extends {
    [key: string]: ExecutableStep;
}> {
    results: Results<TSteps>;
}
/**
 * A plan that represents an object using the keys given and the values being
 * the results of the associated plans.
 */
export declare class ObjectStep<TPlans extends {
    [key: string]: ExecutableStep;
} = {
    [key: string]: ExecutableStep;
}> extends UnbatchedExecutableStep<DataFromObjectSteps<TPlans>> implements SetterCapableStep<TPlans> {
    static $$export: {
        moduleName: string;
        exportName: string;
    };
    isSyncAndSafe: boolean;
    allowMultipleOptimizations: boolean;
    private keys;
    optimizeMetaKey: string;
    constructor(obj: TPlans);
    /**
     * This key doesn't get typed, but it can be added later which can be quite
     * handy.
     */
    set<TKey extends keyof TPlans>(key: TKey, plan: TPlans[TKey]): void;
    getStepForKey<TKey extends keyof TPlans>(key: TKey, allowMissing?: boolean): TKey extends keyof TPlans ? TPlans[TKey] : null;
    toStringMeta(): string;
    tupleToObjectJIT(callback: (fn: (extra: ExecutionExtra, ...tuple: Array<DataFromObjectSteps<TPlans>[keyof TPlans]>) => DataFromObjectSteps<TPlans>) => void): void;
    finalize(): void;
    execute({ indexMap, values, extra, }: ExecutionDetails<Array<DataFromObjectSteps<TPlans>[keyof TPlans]>>): ReadonlyArray<DataFromObjectSteps<TPlans>>;
    unbatchedExecute(_extra: UnbatchedExecutionExtra, ..._values: any[]): any;
    deduplicate(peers: ObjectStep<any>[]): ObjectStep<TPlans>[];
    optimize(opts: StepOptimizeOptions): ConstantStep<any> | this;
    /**
     * Get the original plan with the given key back again.
     */
    get<TKey extends keyof TPlans>(key: TKey): TPlans[TKey];
}
/**
 * A plan that represents an object using the keys given and the values being
 * the results of the associated plans.
 */
export declare function object<TPlans extends {
    [key: string]: ExecutableStep;
}>(obj: TPlans): ObjectStep<TPlans>;
export {};
//# sourceMappingURL=object.d.ts.map